from typing import Any, Union

from miv.core.datatype.events import *
from miv.core.datatype.protocol import *
from miv.core.datatype.pure_python import *
from miv.core.datatype.signal import *
from miv.core.datatype.spikestamps import *

DataTypes = Any  # Union[ # TODO
#    miv.core.datatype.signal.Signal,
#    miv.core.datatype.signal.Spikestamps,
# ]
